/*
 * $Id$
 *
 * Serial Wire Debug Open API.
 * Library Body File.
 *
 * Copyright (C) 2010, Tomasz Boleslaw CEDRO (http://www.tomek.cedro.info)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the Tomasz Boleslaw CEDRO nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.*
 *
 * Written by Tomasz Boleslaw CEDRO <tomek.cedro@gmail.com>, 2010;
 *
 */

#include <libswd.h>
#include <string.h>
#include <stdlib.h>

/* DATA GENERATED BELOW ARE TO BE SHIFTED OUT MSB FIRST, UNLIKE IN JTAG!
 * These functions generate a bistream packets for SWD communications.
 * Bistream is returned in byte array pointed by *buffer parameter.
 * Actual positive byte count is returned by the function.
 * Negative return value means an error. Zero means no data was generated.
 * If you need to swap bit direction, use bitswap functions provided below.
 * All functions name start with "swd_bitgen_" prefix.
 */


/* BitSwap helper function that reverse bit order in *buffer. 
 * Least Significant Bit becomes Most Significant Bit.
 * Rersult contains only "count" bits for use with char or int easily.
 * This function operates on unsigned char (8-bit) data type.
 */
int swd_bitswap8(unsigned char *buffer, int bitcount){
 if (bitcount>8) return SWD_ERROR_BADPARAM;
 unsigned char bit, res=0; //res must be unsigned for proper shifting result
 #ifdef __SWDDEBUG__
 printf("|SWD_DEBUG: swd_bitswap8(%02X, %d);\n", *buffer, bitcount);
 #endif
 for (bit=0;bit<bitcount;bit++) {
  res=(res<<1)|(((*buffer>>bit)&1)?1:0); 
  #ifdef __SWDDEBUG__
  printf("|SWD_DEBUG: swd_bitswap8: in=%02X out=%02X bit=%d\n", *buffer, res, bit);
  #endif
 }
 *buffer=res;
 return SWD_OK;
}

/* BitSwap helper function that reverse bit order in *buffer. 
 * Least Significant Bit becomes Most Significant Bit.
 * Rersult contains only "count" bits for use with char or int easily.
 * This function operates on unsigned int (32-bit) data type.
 */
int swd_bitswap32(unsigned int *buffer, int bitcount){
 if (bitcount>32) return SWD_ERROR_BADPARAM;
 unsigned int bit, res=0; //res must be unsigned for proper shifting result
 #ifdef __SWDDEBUG__
 printf("|SWD_DEBUG: swd_bitswap32(%08X, %d);\n", *buffer, bitcount);
 #endif
 for (bit=0;bit<bitcount;bit++) {
  res=(res<<1)|(((*buffer>>bit)&1)?1:0); 
  #ifdef __SWDDEBUG__
  printf("|SWD_DEBUG: swd_bitswap32: in=%08X out=%08X bit=%d\n", *buffer, res, bit);
  #endif
 }
 *buffer=(int)res;
 return SWD_OK;
}



/* Even parity calculator helper function for int type. */
int swd_paritycheck32(unsigned int data){
 int i,parity=0;
 for (i=0;i<32;i++) parity ^= (data>>i); 
 return parity;
}

int swd_paritycheck8(unsigned char data){
 char i,parity=0;
 for (i=0;i<8;i++) parity ^= (data>>i);
 return parity;
}




/* Command Queue helper functions.
 * All functions start with "swd_cmd_queue_" prefix. 
 */

/* Find first element of a queue pointed by *cmdq element. */
swd_cmd_t* swd_cmd_queue_find_root(swd_cmd_t *cmdq){
 if (cmdq==NULL) return NULL;
 swd_cmd_t *cmd=cmdq;
 while (cmd->prev!=NULL) cmd=cmd->prev;
 return cmd;
}

/* Find last element of a queue pointed by *cmdq element. */
swd_cmd_t* swd_cmd_queue_find_end(swd_cmd_t *cmdq){
 if (cmdq==NULL) return NULL;
 swd_cmd_t *cmd=cmdq;
 while (cmd->next!=NULL) cmd=cmd->next;
 return cmd;
}

/* Append element pointed by *cmd at the end of the quque pointed by *cmdq. */
int swd_cmd_queue_append(swd_cmd_t *cmdq, swd_cmd_t *cmd){
 if (cmdq==NULL || cmd==NULL) return SWD_ERROR_NULLPOINTER;
 if (cmdq->next != NULL){
  swd_cmd_t *lastcmd;
  lastcmd=swd_cmd_queue_find_end(cmdq);
  lastcmd->next=cmd;
  cmd->prev=lastcmd;
 } else {
  cmdq->next=cmd;
  cmd->prev=cmdq;
 }
 cmdq=cmd; 
 return 1;
}

/* Free memory / destroy queue pointed by the *cmdq element. */
int swd_cmd_queue_free(swd_cmd_t *cmdq){
 swd_cmd_t *cmd, *nextcmd;
 cmd=swd_cmd_queue_find_root(cmdq);
 while (cmd!=NULL) {
  nextcmd=cmd->next;
  free(cmd);
  cmd=nextcmd;
 }
 return SWD_OK;
}

/* Free memory / destroy / shorten queue head from root to the *cmdq element. */
int swd_cmd_queue_free_head(swd_cmd_t *cmdq){
 swd_cmd_t *cmdqroot, *nextcmd;
 cmdqroot=swd_cmd_queue_find_root(cmdq);
 while(cmdqroot!=cmdq){
  nextcmd=cmdqroot->next;
  free(cmdqroot);
  cmdqroot=nextcmd;
 }
 cmdqroot->prev=NULL;
 return SWD_OK;
}

int swd_cmd_queue_free_tail(swd_cmd_t *cmdq){
 swd_cmd_t *cmdqend, *nextcmd;
 nextcmd=cmdq;
 cmdqend=swd_cmd_queue_find_end(cmdq);
 if (cmdqend==NULL) return SWD_ERROR_QUEUE; 
 while(nextcmd!=cmdq){
  nextcmd=cmdqend->prev;
  free(cmdqend);
  cmdqend=nextcmd;
 }
 cmdqend->next=NULL;
 return SWD_OK;
}





/* SWD Commands Queue Elements Generator Routines.
 * These command quants are created in memory and can be easily appended
 * to the queue pointed by the *cmdq element.
 * All functions start with "swd_cmd_append_" prefix. */

/* Append SWD Request Packet Header to the context command queue.
 * Note that request is not validated here so bad request can be sent as well.*/
int swd_cmd_append_request(swd_ctx_t *swdctx, char request){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 swd_cmd_t *cmd;
 cmd=calloc(1,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd->request=request;
 cmd->bits=SWD_REQUEST_BITLEN;
 cmd->type=SWD_CMDTYPE_MOSI;
 return swd_cmd_queue_append(swdctx->cmdq, cmd);
} 

/* Append SWD Turnaround to the context command queue. */ 
int swd_cmd_append_trn(swd_ctx_t *swdctx, int RnW){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
  swd_cmd_t *cmd;
  cmd=calloc(1,sizeof(swd_cmd_t));
  if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
  cmd->RnW=RnW;
  cmd->bits=swdctx->params.trnlen;
  cmd->type=SWD_CMDTYPE_TRN;
  return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

/* Append bus binary read bit-by-bit operation into context command queue.
 * Note that each command returns binary value into integer data element.
 * On error memory is freed and error code is returned.
 * Memory pointed by *data must be allocated prior calling this function. */
int swd_cmd_append_bitread(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<=0) return SWD_ERROR_BADPARAM;
 swd_cmd_t **cmd, *tmpcmdq;
 cmd=calloc(count,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 int i;
 for (i=0;i<count;i++){
  cmd[i]->bits=1;
  cmd[i]->type=SWD_CMDTYPE_BITMISO;
  if (swd_cmd_queue_append(swdctx->cmdq, cmd[i])!=1){
   swd_cmd_queue_free_tail(cmd[0]); // if that fails, just clean the memory
   for (i=0;i<count;i++) free(cmd[i]);
   return SWD_ERROR_QUEUE;
  }
 }
 return count;
}


/* Append 32-bit element + 1-bit parity read operation into context cmd queue.*/
int swd_cmd_append_read(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL || data==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t **cmd;
 int ret;
 cmd=calloc(2,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 //?cmd[0]->misodata=&data;
 cmd[0]->bits=32;
 cmd[0]->type=SWD_CMDTYPE_MISO;
 cmd[1]->parity;
 cmd[1]->bits=1;
 cmd[1]->type=SWD_CMDTYPE_PARMISO;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[0]);
 if (ret<0) return ret;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[1]);
 if (ret<0) return ret;
 return 2;
}

/* Append n-element read operation into context command queue. */
int swd_cmd_append_nread(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<=0) return SWD_ERROR_BADPARAM;
 int i;
 for (i=0;i<=count;i++) if (swd_cmd_append_read(swdctx, &data[i])<1) return SWD_ERROR_BADRESULT;
 return i;
}


int swd_cmd_append_write(swd_ctx_t *swdctx, int *data){
 if (swdctx==NULL || data==NULL) return SWD_ERROR_NULLPOINTER;
 swd_cmd_t **cmd;
 int ret;
 cmd=calloc(2,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 cmd[0]->mosidata=*data;
 cmd[0]->bits=32;
 cmd[0]->type=SWD_CMDTYPE_MOSI;
 cmd[1]->parity=swd_parity_check32(*data);
 cmd[1]->bits=1;
 cmd[1]->type=SWD_CMDTYPE_MOSI;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[0]);
 if (ret<0) return ret;
 ret=swd_cmd_queue_append(swdctx->cmdq, cmd[1]);
 if (ret<0) return ret;
 return 2;
}

int swd_cmd_append_nwrite(swd_ctx_t *swdctx, int *data, int count){
 if (swdctx==NULL || data==NULL || count<1) return SWD_ERROR_BADPARAM;
 int i;
 for (i=0;i<count;i++) if (swd_cmd_append_write(swdctx, &data[i])<1) return SWD_ERROR_BADRESULT;
 return i;
}

/* Append Acnkowledge request from slave to the command queue. */
int swd_cmd_append_ack(swd_ctx_t *swdctx){
 if (swdctx==NULL) return SWD_ERROR_NULLCONTEXT;
 swd_cmd_t *cmd;
 cmd=calloc(1,sizeof(swd_cmd_t)); 
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;  
 cmd->bits=SWD_ACK_BITLEN;
 cmd->type=SWD_CMDTYPE_ACK;
 return swd_cmd_queue_append(swdctx->cmdq, cmd);
}

/* Append n-octet size control sequence to the context command queue. */
int swd_cmd_append_control(swd_ctx_t *swdctx, char *ctlmsg, int len){
 int elm, ret;
 swd_cmd_t **cmd;
 if (len<=0) return SWD_ERROR_BADPARAM;
 cmd=calloc(len,sizeof(swd_cmd_t));
 if (cmd==NULL) return SWD_ERROR_OUTOFMEM;
 for (elm=0;elm<len;elm++){
  cmd[elm]->control=ctlmsg[elm];
  cmd[elm]->type=SWD_CMDTYPE_CONTROL;
  cmd[elm]->bits=sizeof(ctlmsg[elm]);
  ret=swd_cmd_queue_append(swdctx->cmdq, cmd[elm]); 
  if (ret<=1) return ret;
 }
 return elm;
}

int swd_cmd_append_swdpreset(swd_ctx_t *swdctx){
 return swd_cmd_append_control(swdctx, SWD_CMD_SWDPRESET, sizeof(SWD_CMD_SWDPRESET));
}

int swd_cmd_append_jtag2swd(swd_ctx_t *swdctx){
 return swd_cmd_append_control(swdctx, SWD_CMD_JTAG2SWD, sizeof(SWD_CMD_JTAG2SWD));
}

int swd_cmd_append_swd2jtag(swd_ctx_t *swdctx){
 return swd_cmd_append_control(swdctx, SWD_CMD_SWD2JTAG, sizeof(SWD_CMD_SWD2JTAG));
}



